"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
class YamprintFormatter {
    constructor() {
        this.circular = "~Circular~";
        this.arrayPrefix = "► ";
        this.indent = "  ";
        this.multilineMargin = "| ";
        this.undefined = "undefined";
        this.nul = "null";
    }
    _formatCtorTag(line) {
        return line && `|${line}|` || "";
    }
    scalarObjectTag(obj) {
        if (obj.constructor === Object)
            return "{}";
        if (!obj.constructor || !obj.constructor.name)
            return this._formatCtorTag("~anonymous~");
        return this.constructorTag(obj);
    }
    propertyKey(key) {
        key = /[^a-zA-Z0-9_$]/.test(key) ? `'${key}'` : key;
        return `${key} = `;
    }
    theme(theme) {
        let self = this;
        class ThemedYampritFormatter extends YamprintFormatter {
            constructor(colors) {
                super();
                Object.keys(colors).forEach(k => {
                    if (typeof this[k] == "function") {
                        this[k] = function (...args) {
                            return colors[k].call(colors, self[k].apply(this, args));
                        };
                    }
                    else {
                        this[k] = colors[k].call(colors, this[k]);
                    }
                });
            }
        }
        return new ThemedYampritFormatter(theme);
    }
    constructorTag(instance) {
        let line = "";
        if (instance instanceof Error) {
            let regularName = instance.name;
            if (regularName === "Error" || !regularName) {
                regularName = instance.constructor && instance.constructor.name;
            }
            regularName = regularName || "Error";
            line += regularName;
            if (instance.message) {
                line += `("${instance.message}")`;
            }
        }
        else {
            let ctor = instance.constructor;
            if (ctor === Object || !ctor) {
                return "";
            }
            if (instance.constructor) {
                line = instance.constructor.name || "~anonymous~";
            }
        }
        return this._formatCtorTag(line);
    }
    ;
    sparseArrayIndex(ix) {
        return !Number.isNaN(+ix) ? `(${ix}) ` : `'${ix}' `;
    }
    function(f) {
        let line = "function";
        if (f.name) {
            line += " " + f.name;
        }
        return `|${line}|`;
    }
    threwAlert(err) {
        let line = "THREW";
        if (!(err instanceof Error)) {
            line += " " + (err && err.toString || Object.prototype.toString).call(err);
        }
        else {
            if (err.name) {
                line += " " + err.name;
            }
            else if (err.constructor && err.constructor.name) {
                line += `${err.constructor.name}`;
            }
            else {
                line += " Error";
            }
            if (err.message) {
                line += `("${err.message}")`;
            }
        }
        return `‼ ${line} ‼`;
    }
    number(n) {
        return n.toString();
    }
    ;
    string(s) {
        if (!s.match(/[(\r|\n)]/)) {
            return `'${s}'`;
        }
        else {
            return s;
        }
    }
    symbol(s) {
        return s.toString();
    }
    boolean(b) {
        return b.toString();
    }
    date(d) {
        return d.toString();
    }
    regexp(r) {
        return r.toString();
    }
    fallback(o) {
        let proto = Object.getPrototypeOf(o);
        if (Array.isArray(o) && Object.keys(o).length === 0) {
            return "[]";
        }
        return null;
    }
    formatScalar(value) {
        if (value === null) {
            return this.nul;
        }
        if (value === undefined) {
            return this.undefined;
        }
        if ([String, Number, Boolean, Symbol, Function].some(ctor => value instanceof ctor)) {
            value = value.valueOf();
        }
        switch (typeof value) {
            case "boolean":
                return this.boolean(value);
            case "string":
                return this.string(value);
            case "symbol":
                return this.symbol(value);
            case "number":
                return this.number(value);
            case "function":
                return this.function(value);
        }
        if (value instanceof Date) {
            return this.date(value);
        }
        else if (value instanceof RegExp) {
            return this.regexp(value);
        }
        else {
            return this.fallback(value);
        }
    }
}
exports.YamprintFormatter = YamprintFormatter;

//# sourceMappingURL=yamprint-formatter.js.map
